#include <stdio.h>
#include <io.h>
#include <string.h>
#include <setjmp.h>
#include <sys/stat.h>

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>

#include "common.h"
#include "locodef.h"
#include "dump.h"
#include "xml.h"

int embedpng = 0;
int allinone = 0;
char *filebase;

__declspec(dllexport) BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdReason, LPVOID lpvReserved)
{
  return TRUE;
}

void setfilebase(char *fname)
{

  filebase = my_alloc(strlen(fname)+1, "filebase");

  strcpy(filebase, fname);
  if (strrchr(filebase, '.'))
	strrchr(filebase, '.')[0] = 0;
}

uint8_t *rledecode(uint8_t *data, uint32_t len)
{
  uint8_t *chunk;
  uint32_t chunklen, chunkofs;

  chunklen = len*2;
  chunk = my_alloc(len*2, "chunk");

  chunkofs = 0;
  while (len>0) {
	int8_t rle;
	int run;

	rle = *(int8_t*) data++; len--;
	run = abs(rle)+1;

	if (chunkofs+run > chunklen) {
		uint32_t oldlen = chunklen;
		while (chunkofs+run > chunklen) 
			chunklen <<= 1;
		chunk = my_realloc(chunk, chunklen, oldlen, "chunk");
	}

	if (rle<0) {
		int i;
		uint8_t byte = *data++; len--;

		memset(chunk+chunkofs, byte, run);
	} else {
		memcpy(chunk+chunkofs, data, run);
		len-=run;
		data+=run;
	}
	chunkofs += run;
  }

  chunk = my_realloc(chunk, chunkofs, chunklen, "chunk");

  return chunk;
}

int rleencode(uint8_t *data, uint32_t len, uint8_t **rleout, uint32_t *rlelen)
{
  uint8_t *out;
  uint32_t start, ofs, outlen, outofs;
  int run;
  uint8_t byte, rle;

  outlen = len*2;
  out = my_alloc(outlen, "out");

  ofs = 0;
  outofs = 0;
  while (len) {
	start = ofs;

	byte = data[ofs++];
	len--;

	run = 1;
	if (len && (byte == data[ofs])) {
		while (len && (byte == data[ofs]) && (run < 127)) {
			run++;
			ofs++;
			len--;
		}
		rle = -(run-1);
		out[outofs++] = rle;
		out[outofs++] = byte;
	} else {
		while (len && (byte != data[ofs]) && (run < 126)) {
			byte = data[ofs];
			run++;
			ofs++;
			len--;
		}
		if (len) {
			run--;
			ofs--;
			len++;
		}
		rle = run-1;
		out[outofs++] = rle;
		memcpy(out+outofs, data+start, run);
		outofs += run;
	}
  }
  *rlelen = outofs;
  *rleout = out;
  return 1;
}

// define a bit rotate (gcc optimizes this into an i386 ROL instruction)
#define _rotl32(a,n) (((a)<<(n))|((a)>>(32-n)))

uint32_t checksum(uint8_t *data, uint32_t len)
{
  uint32_t i, c;

  c = 0xF369A75B;

  c = _rotl32(c ^ data[0], 11);
  for (i=4; i<12; i++)
	c = _rotl32(c ^ data[i], 11);
  for (i=16; i<len; i++)
	c = _rotl32(c ^ data[i], 11);

  return c;
}

__declspec(dllexport) int decode(char *datname)
{
  int class;
  uint32_t subclass;
  uint8_t *chunk;
  uint32_t dumped;

  uint8_t *data, *ptr, *end;
  struct stat st;
  FILE *dat, *xml;
  char *xmlname;
  jmp_buf decode_err;

  err_return = &decode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(datname);

  xmlname = my_alloc(strlen(filebase)+5, "xmlname");
  strcpy(xmlname, filebase);
  strcat(xmlname, ".xml");

  dat = fopen(datname, "rb");
  if (!dat) die("Can't read %s: %s", datname, strerror(errno));

  if (fstat(fileno(dat), &st)) die("Can't stat %s: %s", datname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  xml = fopen(xmlname, "wt");
  if (!xml) die("Can't write %s: %s", xmlname, strerror(errno));
  my_free(xmlname);

  data = my_alloc(st.st_size, "data");

  fread(data, 1, st.st_size, dat);

  if (data[3] != 0x11) die("File %s is not a .dat file", datname);
  end = data+st.st_size;

  fputs("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n", xml);

  class = data[0];
  subclass = getvalue(data+1, 3);

  fprintf(xml, "<object class=\"0x%02X\" subclass=\"0x%06X\" name=\"",
	class, subclass);
  putxmlstr(xml, data+4, 8);
  fputs("\">", xml);

  ptr = data+16;

  class &= 0x7f;

  while (ptr < end) {
	int code;
	uint32_t len;

	code = *ptr; ptr++;
	len = *(uint32_t*) ptr; ptr+=4;

	fprintf(xml, "<chunk compression=\"%d\">\n", code);
	switch(code) {
		case 0:
			dumpchunk(xml, ptr, class);
			break;
		case 1:
			chunk = rledecode(ptr, len);
			dumpchunk(xml, chunk, class);
			my_free(chunk);
			break;
		default:
			die("Can't decode chunk type %d", code);
	}

	fputs("</chunk>", xml);
	ptr += len;
  }
  fputs("</object>\n", xml);

  fclose(xml);

  my_free(data);
  my_free(filebase);

  err_return = NULL;
  return 1;
}

__declspec(dllexport) int encode(char *xmlname)
{
  uint32_t ofs, rlelen, chksum, encsize;
  struct stat st;

  FILE *xml, *dat;
  char *datname, *tmpname, *bakname;
  uint8_t *xmldata, *rle, *encdata;
  jmp_buf encode_err;

  err_return = &encode_err;
  if (setjmp(*err_return)) {
	err_return = NULL;
	return 0;
  }

  setfilebase(xmlname);

  datname = my_alloc(strlen(filebase)+5, "datname");
  tmpname = my_alloc(strlen(filebase)+5, "tmpname");
  bakname = my_alloc(strlen(filebase)+5, "bakname");
  strcpy(datname, filebase);
  strcat(datname, ".dat");
  strcpy(tmpname, filebase);
  strcat(tmpname, ".tmp");
  strcpy(bakname, filebase);
  strcat(bakname, ".bak");

  xml = fopen(xmlname, "rb");	// rt for real xml!!
  if (!xml) die("Can't read %s: %s", xmlname, strerror(errno));

  if (fstat(fileno(xml), &st)) die("Can't stat %s: %s", xmlname, strerror(errno));

#if defined(__BORLANDC__) && __BORLANDC__ < 0x500
  if (st.st_size > 65000) st.st_size = 65000;
#endif

  xmldata = my_alloc(st.st_size, "data");
  fread(xmldata, 1, st.st_size, xml);
  fclose(xml);

  encsize = parsexml(xmldata, st.st_size, &encdata);

  dat = fopen(tmpname, "w+b");
  fwrite(encdata, 1, 12, dat);
  chksum = checksum(encdata, encsize);
  fwrite(&chksum, 4, 1, dat);

  rleencode(encdata+16, encsize-16, &rle, &rlelen);

  encdata[0] = 1;		// chunk type rle encoded
  fwrite(encdata, 1, 1, dat);

  fwrite(&rlelen, 4, 1, dat);
  fwrite(rle, 1, rlelen, dat);

  fclose(dat);

  if (!access(datname, 0)) {	// does datname exist?
	if (access(bakname, 0)) {
		rename(datname, bakname);
	} else {
		remove(datname);
	}
  }
  rename(tmpname, datname);

  my_free(rle);
  my_free(encdata);
  my_free(xmldata);
  my_free(filebase);

  err_return = NULL;
  return 1;
}
