#ifndef _LOCODEF_H
#define _LOCODEF_H

#include <stdio.h>

/* Global variables */


extern int embedpng;
extern int allinone;
extern char *filebase;

extern const char *spriteflags[];
 

enum en_elemid {
	el_bit, el_bitmask, el_cargo, el_cargotype, el_chunk,
	el_description, el_object, el_pngfile, el_sprite,
	el_structure, el_unknown, el_useobject, el_variable,

	el_document,
};
typedef enum en_elemid elemid;

enum en_attrid {
	at_capacity, at_class, at_compression, at_id,
	at_language, at_num, at_size, at_subclass, at_xofs, at_yofs,

	at_STRINGVALUE,

	at_name, at_desc,

	at_LASTVALUE,
};
typedef enum en_attrid attrid;

extern const char *attname[at_LASTVALUE];


enum en_useobjid {
	ob_tracktype, ob_trackmod, ob_visualeffect, ob_wakeeffect,
	ob_rackrail, ob_compatible, ob_startsnd, ob_soundeffect,

	ob_LASTVALUE,
};
typedef enum en_useobjid useobjid;

extern const char *useobjnames[ob_LASTVALUE];


struct s_varinf {
	int ofs;
	int size;
	int num;
	const char *name;
	struct s_varinf *structvars;
	const char **flags;
};
typedef struct s_varinf varinf;

struct s_spriteinf {
	uint32_t width,height,flags, spritesize;
	int32_t xofs,yofs;
};
typedef struct s_spriteinf spriteinf;

struct s_useobjinf {
	int class, num;
	useobjid type;
	uint8_t obj[16];
};
typedef struct s_useobjinf useobjinf;

union u_attval {
	uint32_t u32;
	char *str;
};
typedef union u_attval attval;

union u_elemdesc {
	const char **flags;
	varinf *vars;
};
typedef union u_elemdesc elemdesc;

struct s_eleminf {
	elemid id;
	attval atts[at_LASTVALUE];
	elemdesc desc;
	void *data;
};
typedef struct s_eleminf eleminf;

struct s_arrayinf {
	uint32_t size, ent, entsize;
	char ***data;
};
typedef struct s_arrayinf arrayinf;

struct s_xmlinf {
	int chardata;	// do we want character data?

	int class, compression;
	uint32_t subclass;
        uint8_t name[8];

	uint32_t curvaluelen, curvaluesize;
	char *curvalue;

	int curelemind;
	const char **curatts;

	arrayinf objdata_i;
	uint8_t *objdata;

	arrayinf descdata_i;
	uint8_t *descdata;

	arrayinf useobj_i;
	useobjinf **useobj;

        uint32_t capsize[2], caplen[2];
	uint8_t *caps[2];

	arrayinf sprites_i;
	spriteinf **sprites;

        uint32_t cursprite;
	arrayinf spritedata_i;
	uint8_t **spritedata;

	arrayinf elems_i;
	eleminf **elems;
};
typedef struct s_xmlinf xmlinf;


typedef uint32_t dumpproc(FILE*, uint8_t*);
typedef void parseproc(xmlinf*);

struct s_objtype {
	varinf *vars;
	uint32_t size;
	dumpproc *dump;
	parseproc *parse;
};
typedef struct s_objtype objtype;

#define NUMOBJTYPES 0x22
extern objtype objtypes[NUMOBJTYPES];


#endif // ndef _LOCODEF_H
