#include <stdio.h>
#include <string.h>
#include <expat.h>

#include "common.h"
#include "locodef.h"
#include "png.h"

void putxmlstr(FILE *f, const char *buffer, uint32_t minlen)
{
  uint32_t i;

  if (!minlen) minlen = strlen(buffer);

  for (i=0; i<minlen; i++)
	switch (buffer[i]) {
		case '"': fputs("&quot;", f); break;
		case '<': fputs("&lt;", f); break;
		case '>': fputs("&gt;", f); break;
		case '&': fputs("&amp;", f); break;
//		case 0 ... 31: fprintf(f, "&#x%02X;", buffer[i]); break;
		default: fputc(buffer[i], f); break;
	}
}

uint32_t checklen(arrayinf *inf, int32_t ofs)
{
  uint32_t oldsize = inf->size;

  if (ofs<0) ofs=inf->ent;
  if (ofs>=inf->ent) inf->ent=ofs+1;

  if (inf->size >= (ofs+1)*inf->entsize) return ofs;

  while (inf->size < (ofs+1)*inf->entsize)
	inf->size += 128;

  if (oldsize)
	*(inf->data) = my_realloc(*(inf->data), inf->size, oldsize, "inf->data");
  else
	*(inf->data) = my_alloc(inf->size, "inf->data");

  return ofs;
}

void addarrent(arrayinf *inf, void *ent)
{
  uint32_t ofs = checklen(inf, -1);
  if ((*inf->data)[ofs]) my_free((*inf->data)[ofs]);
  (*inf->data)[ofs] = ent;
}

void setarrent(arrayinf *inf, int32_t ofs, void *ent)
{
  checklen(inf, ofs);
  if ((*inf->data)[ofs]) my_free((*inf->data)[ofs]);
  (*inf->data)[ofs] = ent;
}

typedef void elemproc(xmlinf *inf, elemid id, int depth);

struct s_elem {
	const char *name;
	elemid id, container1, container2;
	elemproc *start, *end;
	int haschardata;
};
typedef struct s_elem elem;
elem elemlist[];


void parsename(char **name, int *field, int *arrind)
{
  char *f, *a;

  *arrind = -1;
  *field = -1;

  a = strchr(*name, '[');
  f = strchr(*name, '_');

  if (a) {
	a[0] = 0;
	*arrind = strtol(a+1, NULL, 0);
  };
  if (f) {
	f[0] = 0;
	*field = strtol(f+1, NULL, 16);
  }
}

void findvar(xmlinf *inf, int depth, attrid attr, eleminf **elem, varinf **vars,
		int *field, int *num)
{
  char *name;
  int i;

  *elem = inf->elems[depth];
  name = (*elem)->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, field, num);

  *vars = inf->elems[depth-1]->desc.vars;

  if (*field != -1) return;

  for (i=0; (*vars)[i].ofs != -1; i++)
	if (!strcmp(name, (*vars)[i].name)) {
		*field = (*vars)[i].ofs;
		return;
	}

  die("Unknown variable name %s", name);
}

int findbit(xmlinf *inf, int depth, attrid attr, eleminf **elem)
{
  char *name;
  int id, ind;
  varinf *vars;

  *elem = inf->elems[depth-1];
  name = inf->elems[depth]->atts[attr].str;
  if (!name) die("Element has no name!");

  parsename(&name, &id, &ind);

  if (id != -1) return id;

  if (!(*elem)->desc.flags)
	die("No bit flag information available");

  for (id=0; (*elem)->desc.flags[id]; id++)
	if (!strcmp((*elem)->desc.flags[id], name))
		return id;

  die("Unknown bit name %s", name);
  return 0;	// to make the compiler happy
}

void setdata(eleminf *elem, eleminf *parent, int size, int ofs, int num)
{
  elem->data = ((uint8_t*)parent->data)+ofs+size*num;
}

void st_null(xmlinf *inf, elemid id, int depth)
{
}

void en_bit(xmlinf *inf, elemid id, int depth)
{
  int bit, value, size;
  eleminf *elem;

  bit = findbit(inf, depth, at_name, &elem);

  value = strtol(inf->curvalue, NULL, 0);

  size = inf->elems[depth-1]->atts[at_size].u32;
  if (bit > size*8)
	die("Trying to set bit %d beyond size %d", bit, size);

  if (value)
	*(uint32_t*) elem->data |= (1 << bit);
}

void st_bitmask(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var, size, ofs, num;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  if (!size)
	die("Structure '%s' has no size", elem->atts[at_name].str);

  setdata(elem, inf->elems[depth-1], size, ofs, num);

  for (var=0; vars[var].ofs != -1; var++)
	if (vars[var].ofs == ofs)
		elem->desc.flags = vars[var].flags;
  if (!elem->desc.flags)
	die("Unknown bitmask");
}

void st_cargo(xmlinf *inf, elemid id, int depth)
{
  int num;

  num = inf->elems[depth]->atts[at_num].u32;
  if (num>1) die("Invalid capacity number %d", num);

  if (inf->capsize[num]) die("Duplicate capacity number %d", num);

  inf->capsize[num] = 128;
  inf->caps[num] = my_alloc(inf->capsize[num], "caps");
  inf->caps[num][0] = inf->elems[depth]->atts[at_capacity].u32;
  *(uint16_t*) (inf->caps[num]+1) = 0xffff;
  inf->caplen[num] = 3;
}

void en_cargotype(xmlinf *inf, elemid id, int depth)
{
  int num, len, ofs;

  num = inf->elems[depth-1]->atts[at_num].u32;

  ofs = inf->caplen[num];
  inf->caplen[num] += 3;
  while (inf->capsize[num] < inf->caplen[num]) {
	inf->caps[num] = my_realloc(inf->caps[num], inf->capsize[num]+128,
					inf->capsize[num], "caps");
	inf->capsize[num] += 128;
  }

  inf->caps[num][ofs-2] = inf->elems[depth]->atts[at_id].u32;
  *(uint16_t*) (inf->caps[num]+ofs-1) = strtol(inf->curvalue, NULL, 0);
  *(uint16_t*) (inf->caps[num]+ofs+1) = 0xffff;
}

void st_chunk(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem, *parent;

  elem = inf->elems[depth];
  parent = inf->elems[depth-1];

  inf->compression = elem->atts[at_compression].u32;

  elem->data = parent->data;
  elem->desc.flags = parent->desc.flags;
}

void en_description(xmlinf *inf, elemid id, int depth)
{
  int ofs;

  ofs = inf->descdata_i.ent;
  checklen(&inf->descdata_i, ofs+inf->curvaluelen+1);

  inf->descdata[ofs] = inf->elems[depth]->atts[at_language].u32;
  memcpy(inf->descdata+ofs+1, inf->curvalue, inf->curvaluelen);
}

void st_object(xmlinf *inf, elemid id, int depth)
{
  char *name;
  eleminf *elem = inf->elems[depth];

  inf->class = elem->atts[at_class].u32;
  inf->subclass = elem->atts[at_subclass].u32;

  memset(inf->name, ' ', 8);

  name=elem->atts[at_name].str;
  if (strlen(name)>8)
	die("Object name '%s' too long", name);

  memcpy(inf->name, name, strlen(name));	// padded; no terminating zero

  if ((inf->class&0x7f) < NUMOBJTYPES)
	elem->desc.vars = objtypes[inf->class&0x7f].vars;

  if (!elem->desc.vars)
	die("Cannot parse class %X yet", inf->class);

  checklen(&inf->objdata_i, objtypes[inf->class & 0x7f].size);
  elem->data = inf->objdata;
}

void en_pngfile(xmlinf *inf, elemid id, int depth)
{
  uint8_t *spritedata;
  uint32_t width, height, size;
  spriteinf *sprite;

  sprite = inf->sprites[inf->cursprite];

  if (!sprite->flags)
	die("No bits set for sprite id %d", inf->cursprite);

  size = readpng(inf->curvalue, &spritedata, sprite->flags, &width, &height);

  setarrent(&inf->spritedata_i, inf->cursprite, spritedata);
  sprite->width = width;
  sprite->height = height;
  sprite->spritesize = size;

	// prevent <bit> tags after we already have the data
  inf->elems[depth-1]->desc.flags = NULL;
}

void st_sprite(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem = inf->elems[depth];
  spriteinf *sprite = my_alloc(sizeof(spriteinf), "sprite");

  inf->cursprite = elem->atts[at_id].u32;
  elem->data = &sprite->flags;
  elem->desc.flags = spriteflags;
  elem->atts[at_size].u32 = 4;
  sprite->xofs = elem->atts[at_xofs].u32;
  sprite->yofs = elem->atts[at_yofs].u32;

  setarrent(&inf->sprites_i, inf->cursprite, sprite);
}

void en_sprite(xmlinf *inf, elemid id, int depth)
{
  if (!inf->spritedata[inf->cursprite])
	die("No sprite data for sprite id %d", inf->cursprite);
}

void st_structure(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int var, ofs, num, size;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  if (!size)
	die("Structure '%s' has no size", elem->atts[at_name].str);

  setdata(elem, inf->elems[depth-1], size, ofs, num);

  for (var=0; vars[var].ofs != -1; var++)
	if (vars[var].ofs == ofs)
		elem->desc.vars = vars[var].structvars;

  if (!elem->desc.vars)
	die("Structure '%s' has no definition", elem->atts[at_name].str);
}

const char *useobjnames[ob_LASTVALUE] = {
	"tracktype", "trackmod", "visualeffect", "wakeeffect",
	"rackrail", "compatible", "startsnd", "soundeffect",
};

void en_useobject(xmlinf *inf, elemid id, int depth)
{
  useobjid type;
  eleminf *elem;
  char *name;
  int field, num;
  useobjinf *objinf = my_alloc(sizeof(useobjinf), "objinf");

  elem = inf->elems[depth];
  objinf->class = elem->atts[at_class].u32;

  name = elem->atts[at_desc].str;
  parsename(&name, &field, &num);
  if (num == -1) num=0;

  for (type=0; type<ob_LASTVALUE; type++)
	if (!strcmp(name, useobjnames[type]))
		break;
  if (type == ob_LASTVALUE) die("Unknown object %s", elem->atts[at_desc].str);

  objinf->type = type;
  objinf->num = num;

  if (inf->curvaluelen > 8)
	die("Object name '%s' too long", inf->curvalue);

  objinf->obj[0] = objinf->class;
  memset(objinf->obj+4, ' ', 8);
  strncpy(objinf->obj+4, inf->curvalue, 8);

  addarrent(&inf->useobj_i, objinf);
}

void en_variable(xmlinf *inf, elemid id, int depth)
{
  eleminf *elem;
  varinf *vars;
  int ofs, num, size;
  uint32_t value;

  findvar(inf, depth, at_name, &elem, &vars, &ofs, &num);
  if (num == -1) num=0;

  size = elem->atts[at_size].u32;
  setdata(elem, inf->elems[depth-1], size, ofs, num);

  value = strtol(inf->curvalue, NULL, 0);
  switch(size) {
	case 1: *(uint8_t*) elem->data = value; break;
	case 2: *(uint16_t*) elem->data = value; break;
	case 4: *(uint32_t*) elem->data = value; break;
	default: die("Invalid size %d", size);
  }
}

void en_unknown(xmlinf *inf, elemid id, int depth)
{
  en_variable(inf, id, depth);
}

// standard handlers

void en_updepth(xmlinf *inf, elemid id, int depth)
{
}


// define handlers for a single value, a simple container, and a sub-structure
#define XMLVALUE(name,cont1,cont2) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_null, en_ ## name, 1 }
#define XMLCONT(name,cont1,cont2)  \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_ ## name, en_updepth, 0 }
#define XMLSTRUC(name,cont1,cont2) \
	{ #name, el_ ## name, el_ ## cont1, el_ ## cont2, st_ ## name, en_ ## name, 0 }

elem elemlist[] = {
	XMLVALUE(bit, bitmask, sprite),
	XMLCONT(bitmask, chunk, structure),
	XMLCONT(cargo, chunk, chunk),
	XMLVALUE(cargotype, cargo, cargo),
	XMLCONT(chunk, object, object),
	XMLVALUE(description, chunk, chunk),
	XMLCONT(object, document, document),
	XMLVALUE(pngfile, sprite, sprite),
	XMLSTRUC(sprite, chunk, chunk),
	XMLCONT(structure, chunk, structure),
	XMLVALUE(unknown, chunk, structure),
	XMLVALUE(useobject, chunk, chunk),
	XMLVALUE(variable, chunk, structure),
	{ NULL },
};

const char *attname[at_LASTVALUE] = {
	"capacity", "class", "compression", "id",
	"language", "num", "size", "subclass", "xofs", "yofs",

	"",

	"name", "desc",
};


static void XMLCALL startElement(void *userdata, const char *name, const char **atts)
{
  int i, j, cont1, cont2, depth;
  eleminf *elem;
  xmlinf *inf = userdata;

  for (i=0; elemlist[i].name; i++) {
	if (strcmp(name, elemlist[i].name)) continue;

	if (i != elemlist[i].id) die("Inconsistent elemlist entry %s", name);

	depth=inf->elems_i.ent;
	j = inf->elems[depth-1]->id;
	cont1 = elemlist[i].container1;
	cont2 = elemlist[i].container2;
	inf->chardata = elemlist[i].haschardata;

	if ( (j != cont1) && (j != cont2) )
		die("Element %s is inside element %s but can only be a child of %s%s%s",
			name, elemlist[j].name, elemlist[cont1].name,
			cont1 != cont2 ? " or " : "",
			cont1 != cont2 ? elemlist[cont2].name : "");

	elem = my_alloc(sizeof(eleminf), "elem");
	elem->id = i;

	while (*atts) {
		for (j=0; j<=at_LASTVALUE; j++) {
			if (j == at_LASTVALUE)
				die("Unknown attribute '%s' for element '%s'", name, atts[0]);

			if (strcmp(atts[0], attname[j]) == 0) {
				if (j < at_STRINGVALUE)
					elem->atts[j].u32 = strtol(atts[1], NULL, 0);
				else
					elem->atts[j].str = my_strdup(atts[1], "elem atts str");
				atts+=2;
				break;
			}
		}
	}

	addarrent(&inf->elems_i, elem);

	elemlist[i].start(inf, i, depth);
	return;
  }

  die("Unknown XML element %s", name);
}

static void XMLCALL endElement(void *userdata, const char *name)
{
  int i;
  elemid id;
  attrid j;
  eleminf *elem;
  xmlinf *inf = userdata;

  i = inf->elems_i.ent-1;
  elem = inf->elems[i];
  id = elem->id;
  if (strcmp(elemlist[id].name, name))
	die("Incorrect end element %s", name);

  if (inf->chardata && !inf->curvaluelen)
	die("Element '%s' is empty", name);

  elemlist[id].end(inf, id, i);

  for (j=at_STRINGVALUE+1; j<at_LASTVALUE; j++)
	my_free(elem->atts[j].str);

  my_free(inf->elems[i]);
  inf->elems[i]=NULL;
  inf->elems_i.ent--;

  if (inf->curvalue)
	my_free(inf->curvalue);
  inf->curvalue = NULL;
  inf->curvaluelen = 0;
  inf->curvaluesize = 0;
  inf->chardata = 0;
}

static void XMLCALL characterdata(void *userdata, const XML_Char *s, int utf8len)
{
  xmlinf *inf = userdata;
  uint32_t i, len, newlen, oldlen, newsize;
  uint8_t c;

  if (!inf->chardata) return;

  len = 0;
  for (i=0; i<utf8len; i++) {
	c = s[i];
	if ((c>0xe0) || ( (c>=0x80) && (c<0xc2) ) )
		die("Invalid character sequence %02X%02X", (uint8_t) s[i], (uint8_t) s[i+1]);
	if (c & 0x80) {
		c &= 0xf0;
		while (c & 0x80) {
			c <<= 1;
			i++;
		}
		i--;
	}
	len++;
  }

  oldlen = inf->curvaluelen;
  newlen = oldlen + (uint32_t) len;
  newsize = inf->curvaluesize;

  if (newsize <= newlen+1) {
	while (newsize <= newlen+1)
		newsize += 128;

	if (oldlen)
		inf->curvalue = my_realloc(inf->curvalue, newsize, oldlen, "curvalue");
	else
		inf->curvalue = my_alloc(newlen + 1, "curvalue");
	inf->curvaluesize = newsize;
  }

  len = 0;
  for (i=0; i<utf8len; ) {
	c = s[i++];
	if (c & 0x80)
		c = ((c & 0x3f) << 6) + (s[i++] & 0x3f);
	inf->curvalue[len++] = c;
  }
  inf->curvalue[oldlen+len] = 0;
  inf->curvaluelen=newlen;
}


void addval(xmlinf *inf, uint8_t *val, uint32_t len)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+len-1);
  memcpy(inf->objdata+ofs, val, len);
}

void add32(xmlinf *inf, uint32_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+3);
  *(uint32_t*) (inf->objdata+ofs) = val;
}

void add16(xmlinf *inf, uint16_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs+1);
  *(uint16_t*) (inf->objdata+ofs) = val;
}

void add8(xmlinf *inf, uint8_t val)
{
  uint32_t ofs = inf->objdata_i.ent;

  checklen(&inf->objdata_i, ofs);
  inf->objdata[ofs] = val;
}


void addobj(xmlinf *inf, useobjid id, int num)
{
  int i;

  for (i=0; i<inf->useobj_i.ent; i++)
	if ( (inf->useobj[i]->type == id) && (inf->useobj[i]->num == num) ) {
		addval(inf, inf->useobj[i]->obj, 16);
		return;
	}

  die("Could not find object to use for %s[%d]", useobjnames[id], num);
}

void addsprites(xmlinf *inf)
{
  uint32_t i, size, ofs;

  add32(inf, inf->sprites_i.ent);

  size = 0;
  for (i=0; i<inf->sprites_i.ent; i++)
	size += inf->sprites[i]->spritesize;

  add32(inf, size);

  ofs = 0;
  for (i=0; i<inf->sprites_i.ent; i++) {
	spriteinf *s=inf->sprites[i];

	add32(inf, ofs);
	add16(inf, s->width);
	add16(inf, s->height);
	add16(inf, s->xofs);
	add16(inf, s->yofs);
	add32(inf, s->flags);

	ofs += inf->sprites[i]->spritesize;
  }

  for (i=0; i<inf->sprites_i.ent; i++)
	addval(inf, inf->spritedata[i], inf->sprites[i]->spritesize);
}


#define MAKEARR(name) \
	parsedinfo->name ## _i.entsize = sizeof(parsedinfo->name[0]); \
	parsedinfo->name ## _i.data = (void*) &(parsedinfo->name);

void freearr(arrayinf *inf)
{
  int i;

  for (i=0; i<inf->ent; i++)
	my_free((*inf->data)[i]);
  my_free(*inf->data);
  *inf->data=NULL;
}

uint32_t parsexml(uint8_t *data, uint32_t len, uint8_t **outdata)
{
  XML_Parser parser;
  xmlinf *parsedinfo;
  eleminf *doc;
  uint32_t outlen;
  int i;

  parsedinfo = my_alloc(sizeof(xmlinf), "parsedinfo");
  MAKEARR(objdata);
  MAKEARR(descdata);
  MAKEARR(useobj);
  MAKEARR(sprites);
  MAKEARR(spritedata);
  MAKEARR(elems);

  doc = my_alloc(sizeof(eleminf), "doc");
  doc->id = el_document;
  addarrent(&parsedinfo->elems_i, doc);

  parser = XML_ParserCreate(NULL);
  XML_SetUserData(parser, parsedinfo);
  XML_SetElementHandler(parser, startElement, endElement);
  XML_SetCharacterDataHandler(parser, characterdata);
  if (XML_Parse(parser, data, len, 1) == XML_STATUS_ERROR)
	die("Error while parsing xml: %s at line %d",
	      XML_ErrorString(XML_GetErrorCode(parser)),
	      XML_GetCurrentLineNumber(parser));
  XML_ParserFree(parser);

  if (parsedinfo->elems_i.ent != 1) die("Recursion not terminated");

  checklen(&parsedinfo->descdata_i, parsedinfo->descdata_i.ent);
  parsedinfo->descdata[parsedinfo->descdata_i.ent-1] = 0xff;

  if ((parsedinfo->class & 0x7f) >= NUMOBJTYPES)
	die("This can't happen");

  parsedinfo->objdata_i.ent = objtypes[parsedinfo->class & 0x7f].size;
  objtypes[parsedinfo->class & 0x7f].parse(parsedinfo);

  outlen = parsedinfo->objdata_i.ent;
  *outdata = parsedinfo->objdata;
  checklen(&parsedinfo->objdata_i, outlen+16);

  memmove(parsedinfo->objdata+16, parsedinfo->objdata, outlen);
  outlen += 16;

  parsedinfo->objdata[0] = parsedinfo->class;
  memcpy(parsedinfo->objdata+1, &parsedinfo->subclass, 3);
  memcpy(parsedinfo->objdata+4, parsedinfo->name, 8);
  memset(parsedinfo->objdata+12, 0, 4);

  {
    FILE *f;
    char *dmx = my_alloc(strlen(filebase)+5, "dmx");
    strcpy(dmx, filebase);
    strcat(dmx, ".dmx");
    f=fopen(dmx,"wb"); if (!f) die("write %s: %s", dmx, strerror(errno));
    fwrite(*outdata, 1, outlen, f);
    fclose(f);
    my_free(dmx);
  }

  my_free(parsedinfo->descdata);
  freearr(&parsedinfo->useobj_i);
  my_free(parsedinfo->caps[0]);
  my_free(parsedinfo->caps[1]);
  freearr(&parsedinfo->sprites_i);
  freearr(&parsedinfo->elems_i);
  my_free(parsedinfo);

  return outlen;
}
